#include "triangleMesh.h"
#include "material.h"
#include "RTConfig.h"
#include <qglviewer/vec.h>

#define GL_BGRA 0x80E1
#define GL_TEXTURE_RECTANGLE_EXT 0x84F5

// Pour ne pas avoir a mettre qglviewer:: devant les classes
// definies dans le namespace qglviewer.
using namespace qglviewer;



TriangleMesh::TriangleMesh() : bradius_(1.0f), texID(0) {
}

TriangleMesh::~TriangleMesh() {
	if(triTree)
		delete triTree;
}

void TriangleMesh::draw() const
{
	if(!RTConfig::getInstance()->drawTriangleMesh)
		return;

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  glPushAttrib(GL_ALL_ATTRIB_BITS);

  /*glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDisable(GL_DEPTH_TEST);*/


  glMultMatrixd(frame().matrix());
  glColor3fv(material().diffuseColor());

  float colorMat[] = { material().diffuseColor().r, material().diffuseColor().g, material().diffuseColor().b, 1.0f };
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, colorMat);
  float colorMatSpec[] = { material().specularColor().r, material().specularColor().g, material().specularColor().b, 1.0f };
  glMaterialfv(GL_FRONT, GL_SPECULAR, colorMatSpec);

  glMateriali(GL_FRONT, GL_SHININESS, (int)material().specularCoefficient());



  ////////textures
  if(!this->material().texture().isNull()){
	  if(!this->texID){
		glGenTextures(1, &(this->texID));

		glBindTexture(GL_TEXTURE_2D, this->texID);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, this->material().texture().width(), this->material().texture().height(), 0, GL_BGRA, GL_UNSIGNED_BYTE, this->material().texture().bits() );
	  }

	  glEnable(GL_TEXTURE_2D);
	  glBindTexture(GL_TEXTURE_2D, this->texID);
  }

  ///////
   glColor4f(1.0,1.0,1.0,1.0);

  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  //glEnableClientState(GL_COLOR_ARRAY);

	for (uint i= 0; i<scene3DS.GetMeshCount(); i++) {
        LMesh &mesh = scene3DS.GetMesh(i);

        glVertexPointer(4, GL_FLOAT, 0, &mesh.GetVertex(0));
        glNormalPointer(GL_FLOAT, 0, &mesh.GetNormal(0));
        glColorPointer(3, GL_FLOAT, 0, &mesh.GetBinormal(0));
		glTexCoordPointer (2, GL_FLOAT, 0, &mesh.GetUV(0));
        glDrawElements(GL_TRIANGLES, mesh.GetTriangleCount()*3, 
                        GL_UNSIGNED_SHORT, &mesh.GetTriangle(0));

    }

  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);

   /*glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 // glDisable(GL_DEPTH_TEST);

  glColor4f(0.5,0.0,0.0,0.5);
  GLUquadricObj *quad=gluNewQuadric();
  gluSphere(quad,this->boundingRadius() , 20, 20);*/



  glPopAttrib();
  glPopMatrix();

  if(RTConfig::getInstance()->debugDisplay)
	this->triTree->draw(0);
}

void TriangleMesh::initFromDOMElement(const QDomElement& e)
{
  Object::initFromDOMElement(e);

  if(e.hasAttribute("filename")){
	  QString filename=e.attribute("filename", "");
	  if(!scene3DS.LoadFile(filename.toAscii().data() ))
		  std::cout<<"3DS loading error on file:"<<filename.toAscii().data()<<"\n";

  }

  //Comp bounding volume
  Vec vmin(1000,1000, 1000), vmax(-1000,-1000,-1000);

  for (uint i= 0; i<scene3DS.GetMeshCount(); i++){
		LMesh &mesh = scene3DS.GetMesh(i);
		//pour chaque triangle
		for(int j=0; j<mesh.GetVertexCount(); j++){
			LVector4 ve;
			ve=mesh.GetVertex(j);
			if(ve.x<vmin.x)
				vmin.x=ve.x;
			else if(ve.x>vmax.x)
				vmax.x=ve.x;

			if(ve.y<vmin.y)
				vmin.y=ve.y;
			else if(ve.y>vmax.y)
				vmax.y=ve.y;

			if(ve.z<vmin.z)
				vmin.z=ve.z;
			else if(ve.z>vmax.z)
				vmax.z=ve.z;

		}
  }

  this->bbox[0]=vmin;
  this->bbox[1]=vmax;

  //bradius_=(vmax-vmin).norm()/1.9f;
  bradius_=vmax.norm() > vmin.norm() ? vmax.norm()*0.9 : vmin.norm()*0.9;



  //////////////////////////////////////

  ///Build acceleration tree///
  triTree=new SpacePartition::BoxNode();

  //Pour chaque mesh
  for (uint i= 0; i<scene3DS.GetMeshCount(); i++){
	LMesh &mesh = scene3DS.GetMesh(i);

	
	mesh.Optimize(oFull);

	//pour chaque triangle
	for(int j=0; j<mesh.GetTriangleCount(); j++){
		SpacePartition::Triangle *newtri=new SpacePartition::Triangle();

		LVector4 tr;
		tr=mesh.GetVertex(mesh.GetTriangle(j).a);
		newtri->v[0]=this->frame().inverseCoordinatesOf(Vec(tr.x, tr.y, tr.z));

		tr=mesh.GetVertex(mesh.GetTriangle(j).b);
		newtri->v[1]=this->frame().inverseCoordinatesOf(Vec(tr.x, tr.y, tr.z));

		tr=mesh.GetVertex(mesh.GetTriangle(j).c);
		newtri->v[2]=this->frame().inverseCoordinatesOf(Vec(tr.x, tr.y, tr.z));

		//normals computation, a faire la transfo frame !
		LVector3 n;
		n=mesh.GetNormal(mesh.GetTriangle(j).a);
		newtri->n[0]=Vec(n.x, n.y, n.z);

		n=mesh.GetNormal(mesh.GetTriangle(j).b);
		newtri->n[1]=Vec(n.x, n.y, n.z);

		n=mesh.GetNormal(mesh.GetTriangle(j).c);
		newtri->n[2]=Vec(n.x, n.y, n.z);

		//tangent computation, a faire la transfo frame !
		n=mesh.GetTangent(mesh.GetTriangle(j).a);
		newtri->ts[0][0]=Vec(n.x, n.y, n.z);

		n=mesh.GetTangent(mesh.GetTriangle(j).b);
		newtri->ts[1][0]=Vec(n.x, n.y, n.z);

		n=mesh.GetTangent(mesh.GetTriangle(j).c);
		newtri->ts[2][0]=Vec(n.x, n.y, n.z);

		//Bitangente
		n=mesh.GetBinormal(mesh.GetTriangle(j).a);
		newtri->ts[0][1]=Vec(n.x, n.y, n.z);

		n=mesh.GetBinormal(mesh.GetTriangle(j).b);
		newtri->ts[1][1]=Vec(n.x, n.y, n.z);

		n=mesh.GetBinormal(mesh.GetTriangle(j).c);
		newtri->ts[2][1]=Vec(n.x, n.y, n.z);

		//textures
		LVector2 t;
		t=mesh.GetUV(mesh.GetTriangle(j).a);
		newtri->t[0]=Vec(t.x, t.y, 0);

		t=mesh.GetUV(mesh.GetTriangle(j).b);
		newtri->t[1]=Vec(t.x, t.y, 0);

		t=mesh.GetUV(mesh.GetTriangle(j).c);
		newtri->t[2]=Vec(t.x, t.y, 0);

		//Colors !
		/*LColor3 c;

		//std::cout<<mesh.GetMaterialCount()<<"!\n";
		c=scene3DS.GetMaterial( mesh.GetMaterial(mesh.GetTriangle(j).a)).GetDiffuseColor();
		newtri->c[0]=Vec(c.r, c.g, c.b);

		c=scene3DS.GetMaterial( mesh.GetMaterial(mesh.GetTriangle(j).b)).GetDiffuseColor();
		newtri->c[1]=Vec(c.r, c.g, c.b);

		c=scene3DS.GetMaterial( mesh.GetMaterial(mesh.GetTriangle(j).c)).GetDiffuseColor();
		newtri->c[2]=Vec(c.r, c.g, c.b);*/

		newtri->compBBox();

		triTree->triList.push_back(newtri);
	}
  }

  triTree->compBBox();

  triTree->buildSubNodes(0);

}


bool TriangleMesh::intersect(const Ray &ray, Hit& hit) const{
	bool res=false;

	

#if 0
	if(!bboxIntersect(ray, hit))
		return false;

	//Methode bourine on peut optimiser pour les maillages convexes 
	//avec la dedection de l'entre dans la maillage (arret des tests)

	//Pour chaque mesh
	for (uint i= 0; i<scene3DS.GetMeshCount(); i++){
		LMesh &mesh = scene3DS.GetMesh(i);

		//pour chaque triangle
		for(int j=0; j<mesh.GetTriangleCount(); j++){
			
			Vec p0, p1, p2, normal;
			LVector4 tr;
			tr=mesh.GetVertex(mesh.GetTriangle(j).a);
			p0=Vec(tr.x, tr.y, tr.z);

			tr=mesh.GetVertex(mesh.GetTriangle(j).b);
			p1=Vec(tr.x, tr.y, tr.z);

			tr=mesh.GetVertex(mesh.GetTriangle(j).c);
			p2=Vec(tr.x, tr.y, tr.z);

			LVector3 n;
			n=mesh.GetNormal(mesh.GetTriangle(j).a);

			normal=Vec(n.x, n.y, n.z);
			if( triangleIntersect(ray, hit, p0, p1, p2, normal) );
				res=true;
	
		}

	}
#else

	res=triTree->treeIntersect(ray, hit);

	if(res)
		hit.setMaterial(this->material());
#endif

	return res;
}

